// Dixit - Copyright (C) 2004-2010 Octavian Procopiuc
// Contact: <oprocopiuc@gmail.com>
// Homepage: <http://dixit.sourceforge.net>
//
// Use of this source code is governed by the GPLv3
// license that can be found in the LICENSE.GPL3 file,
// or at <http://www.gnu.org/licenses/gpl.html>.
//
// FILENAME:   mainwindow.cpp
// REVISION:   3
// COMMENTARY: Implements Qt4 MainWindow class
// NAMING:     a_ MainWindow action
//             m_ MainWindow member variable
//             w_ MainWindow member widget

#include "mainwindow.h"
#include "dixit.h"
#include <zlib.h>

// MainWindow's constructor
MainWindow::MainWindow(void)
{
  createStatusBar();
  createMainWindow();
  createToolBar();
  createIndependentActions();
  loadSettings();
  dbStartIndexing();
  return;
}

// MainWindow's destructor
// Frees any allocated resources, no need to delete child widgets, Qt does it all
MainWindow::~MainWindow()
{
  if (m_pThread) {
    m_pThread->wait();
    delete m_pThread;
  }

  if (m_http) {
    delete m_http;
  }

  saveSettings();
  return;
}

// Called from the constructor.
void MainWindow::createStatusBar(void)
{
  statusBar()->showMessage(QString::fromUtf8("Dicționar explicativ al limbii române"));
  return;
}

// Called from the constructor.
void MainWindow::createMainWindow(void)
{
  if (QFile::exists(USER_NEW_IMAGES_PATH + "carte.png")) {
    setWindowIcon(QIcon(USER_NEW_IMAGES_PATH + "carte.png"));

  } else {
    setWindowIcon(QIcon(":/images/book.png"));
  }

  setWindowTitle("Dixit");

  QSplitter *splitter = new QSplitter(this);
  splitter->setOrientation(Qt::Horizontal);

  w_browser = new QTextBrowser(splitter);
  w_browser->setMinimumSize(QSize(200, 0));
  w_browser->setAcceptRichText(true);
  w_browser->setAlignment(Qt::AlignJustify);

  // Set link color to dark blue, etc.
  QPalette pal = w_browser->palette();
  pal.setColor(QPalette::Active,   QPalette::Link, QColor("#003399"));
  pal.setColor(QPalette::Active,   QPalette::Text, QColor("#191919"));
  pal.setColor(QPalette::Inactive, QPalette::Link, QColor("#003399"));
  pal.setColor(QPalette::Inactive, QPalette::Text, QColor("#191919"));
  pal.setColor(QPalette::Disabled, QPalette::Link, QColor("#3366cc"));
  pal.setColor(QPalette::Disabled, QPalette::Text, QColor("#555555"));
  w_browser->setPalette(pal);

  connect(w_browser, SIGNAL(anchorClicked(const QUrl &)),
          this, SLOT(anchorClicked(const QUrl &)));
  connect(w_browser, SIGNAL(highlighted(const QString &)),
          (const QObject *) statusBar(), SLOT(message(const QString &)));
  splitter->addWidget(w_browser);

  w_historyListBox = new QListWidget(splitter);
  w_historyListBox->setViewMode(QListView::ListMode);
  w_historyListBox->setSelectionMode(QAbstractItemView::SingleSelection);
  QSizePolicy sizePolicy((QSizePolicy::Policy)1, (QSizePolicy::Policy)7);
  sizePolicy.setHorizontalStretch(0);
  sizePolicy.setVerticalStretch(0);
  sizePolicy.setHeightForWidth(w_historyListBox->sizePolicy().hasHeightForWidth());
  w_historyListBox->setSizePolicy(sizePolicy);
  w_historyListBox->setMinimumSize(QSize(125, 100));
  w_historyListBox->setMaximumSize(QSize(125, 32767));
  connect(w_historyListBox, SIGNAL(currentTextChanged(const QString &)),
          this, SLOT(wordClickedInHistoryBox(const QString &)));
  splitter->addWidget(w_historyListBox);
  setCentralWidget(splitter);
  return;
}

// Called from the constructor.
void MainWindow::createToolBar(void)
{
  w_toolBar = addToolBar(tr("Dixit"));
  w_toolBar->setEnabled(true);
  w_toolBar->setFixedWidth(730);
  QFont tbFont;
  tbFont.setFamily(tr("Arial"));
  tbFont.setPointSize(10);
  tbFont.setBold(false);
  tbFont.setItalic(false);
  tbFont.setUnderline(false);
  tbFont.setStrikeOut(false);
  w_toolBar->setFont(tbFont);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "inapoi.png")) {
    a_prev = new QAction(QIcon(USER_NEW_IMAGES_PATH + "inapoi.png"),
                         QString::fromUtf8("Înapoi"),
                         this);

  } else {
    a_prev = new QAction(QIcon(":/images/prev.png"),
                         QString::fromUtf8("Înapoi"),
                         this);
  }

  a_prev->setShortcut(tr("Alt+Left"));
  a_prev->setStatusTip(QString::fromUtf8("Înapoi (Alt+Stânga)"));
  a_prev->setEnabled(false);
  connect(a_prev, SIGNAL(triggered()), this, SLOT(backwardInHistory()));
  w_toolBar->addAction(a_prev);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "inainte.png")) {
    a_next = new QAction(QIcon(USER_NEW_IMAGES_PATH + "inainte.png"),
                         QString::fromUtf8("Înainte"),
                         this);

  } else {
    a_next = new QAction(QIcon(":/images/next.png"),
                         QString::fromUtf8("Înainte"),
                         this);
  }

  a_next->setShortcut(tr("Alt+Right"));
  a_next->setStatusTip(QString::fromUtf8("Înainte (Alt+Dreapta)"));
  a_next->setEnabled(false);
  connect(a_next, SIGNAL(triggered()), this, SLOT(forwardInHistory()));
  w_toolBar->addAction(a_next);
  w_toolBar->addSeparator();

  QLabel *searchLabel = new QLabel(QString::fromUtf8("Cu&vântul:"), w_toolBar);
  searchLabel->setMinimumSize(QSize(35, 0));
  QFont lbFont;
  lbFont.setBold(true);
  lbFont.setItalic(false);
  lbFont.setUnderline(false);
  lbFont.setStrikeOut(false);
  searchLabel->setFont(lbFont);

  w_searchBox = new QLineEdit(w_toolBar);
  w_searchBox->setMinimumSize(QSize(175, 0));
  w_searchBox->setMaxLength(50);
  w_searchBox->setFrame(true);

  // Allow only the following on the search line: letters, *, ?, |, -, space.
  QRegExp re("!{0,1}[a-zA-Z*?|^\\[\\]#\\- ]+");
  w_searchBox->setValidator(new QRegExpValidator(re, this));
  connect(w_searchBox, SIGNAL(returnPressed()), this, SLOT(search()));

  // buddies
  searchLabel->setBuddy(w_searchBox);

  w_toolBar->addWidget(searchLabel);
  w_toolBar->addWidget(w_searchBox);

  w_dictionaryComboBox = new QComboBox;
  w_dictionaryComboBox->setEditable(false);
  w_dictionaryComboBox->setMaxVisibleItems(DICTIONARIES_COUNT);
  QFont tbf = w_toolBar->font();
  tbf.setWeight(QFont::Bold);
  w_dictionaryComboBox->setFont(tbf);

  for (int row = 0; row < DICTIONARIES_COUNT; row++) {
    w_dictionaryComboBox->insertItem(row, QString::fromUtf8(DICTIONARIES[row][1]));
  }

  connect(w_dictionaryComboBox, SIGNAL(activated(int)), this, SLOT(setDefaultDictionary(int)));
  connect(w_dictionaryComboBox, SIGNAL(highlighted(int)), this, SLOT(showDictionary(int)));
  w_toolBar->addWidget(w_dictionaryComboBox);

  QAction *action;

  if (QFile::exists(USER_NEW_IMAGES_PATH + "cauta.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "cauta.png"),
                         QString::fromUtf8("Caută"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/find.png"),
                         QString::fromUtf8("Caută"),
                         this);
  }

  action->setShortcut(tr("Alt+C"));
  action->setStatusTip(QString::fromUtf8("Caută (Alt+C)"));
  connect(action, SIGNAL(triggered()), this, SLOT(search()));
  w_toolBar->addAction(action);
  w_toolBar->addSeparator();

  if (QFile::exists(USER_NEW_IMAGES_PATH + "actualizeaza.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "actualizeaza.png"),
                         QString::fromUtf8("Actualizează lista de definiții"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/update.png"),
                         QString::fromUtf8("Actualizează lista de definiții"),
                         this);
  }

  action->setShortcut(tr("Alt+A"));
  action->setStatusTip(QString::fromUtf8("Actualizează lista de definiții (Alt+A)"));
  connect(action, SIGNAL(triggered()), this, SLOT(update()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "tipareste.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "tipareste.png"),
                         QString::fromUtf8("Tipărește"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/print.png"),
                         QString::fromUtf8("Tipărește"),
                         this);
  }

  action->setShortcut(tr("Alt+P"));
  action->setStatusTip(QString::fromUtf8("Tipărește pagina curentă (Alt+P)"));
  connect(action, SIGNAL(triggered()), this, SLOT(print()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "instructiuni.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "instructiuni.png"),
                         QString::fromUtf8("Instrucțiuni"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/help.png"),
                         QString::fromUtf8("Instrucțiuni"),
                         this);
  }

  action->setShortcut(tr("Alt+I"));
  action->setStatusTip(QString::fromUtf8("Instrucțiuni de folosire (Alt+I)"));
  connect(action, SIGNAL(triggered()), this, SLOT(help()));
  w_toolBar->addAction(action);

  if (QFile::exists(USER_NEW_IMAGES_PATH + "iesire.png")) {
    action = new QAction(QIcon(USER_NEW_IMAGES_PATH + "iesire.png"),
                         QString::fromUtf8("Ieșire"),
                         this);

  } else {
    action = new QAction(QIcon(":/images/exit.png"),
                         QString::fromUtf8("Ieșire"),
                         this);
  }

  action->setShortcut(tr("Alt+Q"));
  action->setStatusTip(QString::fromUtf8("Ieșire (Alt+Q)"));
  connect(action, SIGNAL(triggered()), this, SLOT(close()));
  w_toolBar->addAction(action);
  w_toolBar->addSeparator();
  return;
}

// Called from the constructor.
void MainWindow::createIndependentActions(void)
{
  QAction *action;
  action = new QAction(this);
  action->setShortcut(tr("Esc"));
  connect(action, SIGNAL(triggered()), this, SLOT(abortUpdate()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt+F"));
  connect(action, SIGNAL(triggered()), this, SLOT(changeFont()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt++"));
  connect(action, SIGNAL(triggered()), this, SLOT(makeFontBigger()));
  addAction(action);

  action = new QAction(this);
  action->setShortcut(tr("Alt+-"));
  connect(action, SIGNAL(triggered()), this, SLOT(makeFontSmaller()));
  addAction(action);

  m_http = new QHttp(this);
  connect(m_http, SIGNAL(responseHeaderReceived(const QHttpResponseHeader &)),
          this, SLOT(checkHttpResponseHeader(const QHttpResponseHeader &)));
  connect(m_http, SIGNAL(readyRead(const QHttpResponseHeader &)),
          this, SLOT(getDexonlineData(const QHttpResponseHeader &)));
  connect(m_http, SIGNAL(done(bool)), this, SLOT(closeConnection(bool)));
  connect(&m_connectionTimer, SIGNAL(timeout()), this, SLOT(connectionTimeout()));

  return;
}

// Called from the constructor.
void MainWindow::loadSettings(void)
{
#if defined(Q_OS_WIN)
  QString fontFamily = "Arial";
#else
  QString fontFamily = "Sans";
#endif

  QSettings settings(APP_GROUP, APP_KEY);
  QPoint pos          = settings.value("WindowPos", QPoint(100, 100)).toPoint();
  QSize size          = settings.value("WindowSize", QSize(770, 500)).toSize();
  m_autoUpdate        = settings.value("AutoUpdate", bool(false)).toBool();
  m_logInfo           = settings.value("LogInfo", bool(false)).toBool();
  m_defaultDictionary = settings.value("DefaultSource", int(0)).toInt();
  fontFamily          = settings.value("FontFamily", QString(fontFamily)).toString();
  int fontPointSize   = settings.value("FontPointSize", int(12)).toInt();

  w_dictionaryComboBox->setCurrentIndex(m_defaultDictionary);
  resize(size);
  move(pos);

  if (fontPointSize > 16) {
    fontPointSize = 16;

  } else if (fontPointSize < 8) {
    fontPointSize = 8;
  }

  QFont cf = font();
  cf.setFamily(fontFamily);
  cf.setPointSize(fontPointSize);
  cf.setStyleStrategy(QFont::PreferAntialias);
  setFont(cf);
  return;
}

// Called from the destructor.
void MainWindow::saveSettings(void)
{
  QSettings settings(APP_GROUP, APP_KEY);
  settings.setValue("WindowPos", pos());
  settings.setValue("WindowSize", size());
  settings.setValue("AutoUpdate", m_autoUpdate);
  settings.setValue("LogInfo", m_logInfo);
  settings.setValue("DefaultSource", m_defaultDictionary);
  settings.setValue("FontFamily", font().family());
  settings.setValue("FontPointSize", font().pointSize());
  return;
}

QString MainWindow::loadFile(const QString &fileName)
{
  QFile file(fileName);
  file.open(QIODevice::ReadOnly);
  QByteArray htmlData = file.readAll();
  file.close();
  return QString::fromUtf8(htmlData.constData());
}

// This slot is linked to the print button.
void MainWindow::print(void)
{
#ifndef QT_NO_PRINTER
  statusBar()->showMessage(QString::fromUtf8("Tipărește"));
  QTextDocument *document = w_browser->document();
  QPrinter printer;
  QPrintDialog *dlg = new QPrintDialog(&printer, this);

  if (dlg->exec() != QDialog::Accepted) {
    return;
  }

  document->print(&printer);
  statusBar()->showMessage(QString::fromUtf8("Tipărește"), 2000);
#else
  statusBar()->showMessage(QString::fromUtf8("Programul nu poate tipări"), 10000);
  clearBrowser();
  setBrowserHtml(QString::fromUtf8(
                   "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
                   "<font color='#bb3300'><b>Programul nu poate tipări</b></font></td></tr><tr><td><font "
                   "color='#bb3300'>Biblioteca Qt4 nu a fost compilată cu suport pentru tipărire pentru "
                   "platforma dumneavoastră (QT_NO_PRINTER <u>was defined</u>).</font></td></tr></table>"
                 ));
#endif
  return;
}

void MainWindow::processWord(const QString &word)
{
  if (word.isEmpty()) {
    return;
  }

  QString lword(word);
  lword = lword.toLower();

  if (m_strCurrentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_strCurrentWord);
  }

  insertInSearchBox(lword);
  findDefinitions(lword);
  return;
}

// This slot is linked to the help button.
void MainWindow::help(void)
{
  processWord("!instructiuni");
  return;
}

// This slot is linked to w_dictionaryComboBox selection action
void MainWindow::setDefaultDictionary(int id)
{
  m_defaultDictionary = id;
  w_dictionaryComboBox->setCurrentIndex(id);
  return;
}

// This slot is linked to w_dictionaryComboBox selection action
void MainWindow::showDictionary(int id)
{
  statusBar()->showMessage(QString::fromUtf8(DICTIONARIES[id][2]));
  return;
}

// This slot is linked to the input text box and search button.
void MainWindow::search(void)
{
  QString word = w_searchBox->text();

  if (word.isEmpty()) {
    return;
  }

  word = word.toLower();

  if (m_strCurrentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_strCurrentWord);
    m_lstForwardWords.clear();
  }

  findDefinitions(word);
  return;
}

// This slot is linked to the history box item click.
void MainWindow::wordClickedInHistoryBox(const QString &word)
{
  if (word.isEmpty() || word == m_strCurrentWord) {
    return;
  }

  if (m_strCurrentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_strCurrentWord);
    m_lstForwardWords.clear();
  }

  insertInSearchBox(word);
  findDefinitions(word);
  return;
}

// This slot is linked to the w_browser click link.
void MainWindow::anchorClicked(const QUrl &link)
{
  QString word = link.toString();
  processWord(word);
  return;
}

// This slot is linked to the backward button.
void MainWindow::backwardInHistory(void)
{
  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);
    return;
  }

  m_lstForwardWords.push_back(m_strCurrentWord);
  insertInSearchBox(m_lstBackwardWords.last());
  findDefinitions(m_lstBackwardWords.last());
  m_lstBackwardWords.pop_back();

  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);

  } else {
    a_prev->setEnabled(true);
  }

  return;
}

// This slot is linked to the forward button.
void MainWindow::forwardInHistory(void)
{
  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);
    return;
  }

  m_lstBackwardWords.push_back(m_strCurrentWord);
  insertInSearchBox(m_lstForwardWords.last());
  findDefinitions(m_lstForwardWords.last());
  m_lstForwardWords.pop_back();

  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);

  } else {
    a_next->setEnabled(true);
  }

  return;
}

void MainWindow::findDefinitions(const QString &word)
{
  if (word.isEmpty()) {
    return;
  }

  QListWidgetItem *item;
  QList<QListWidgetItem *> foundItems;

  // Limit the history.
  if (w_historyListBox->count() >= MAX_HISTORY_SIZE) {
    item = w_historyListBox->takeItem(w_historyListBox->count() - 1);

    // Items removed from a list widget need to be deleted manually.
    if (item) {
      delete item;
    }
  }

  w_historyListBox->clearSelection();
  foundItems = w_historyListBox->findItems(word, Qt::MatchExactly);

  if (foundItems.count() > 0) { // if item is found, highlight
    item = foundItems.front();
    w_historyListBox->setItemSelected(item, true);

  } else { // else insert word at top of history list.
    w_historyListBox->insertItem(0, word);
    item = w_historyListBox->item(0);
    w_historyListBox->setItemSelected(item, true);
  }

  if (m_lstForwardWords.isEmpty()) {
    a_next->setEnabled(false);

  } else {
    a_next->setEnabled(true);
  }

  if (m_lstBackwardWords.isEmpty()) {
    a_prev->setEnabled(false);

  } else {
    a_prev->setEnabled(true);
  }

  // process all commands
  if ('!' == word[0]) {
    // Check for commands implying QActions.
    if ("!maimare" == word) {
      makeFontBigger();
      return;

    } else if ("!maimic" == word) {
      makeFontSmaller();
      return;
    }

    // Find the anchor for *.html#anchor browser navigation, if applicable
    QString lword(word);
    QString lanchor = lword.section('#', 1);
    lword = lword.section('#', 0, 0);
    clearBrowser();
    // Display html page
    setBrowserHtml(processCommands(lword));

    if (!lanchor.isEmpty()) {
      w_browser->scrollToAnchor(lanchor);
    }

    w_searchBox->selectAll();
    return;
  }

  // search for word definitions in database
  QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
  m_strCurrentWord = word;
  clearBrowser();

  if (m_pThread) {
    m_pThread->wait();
  }

  setBrowserHtml(dbResults(word));
  QApplication::restoreOverrideCursor();

  setFocusToSearchBox();
  return;
}

void MainWindow::setFocusToSearchBox(void)
{
  w_searchBox->setFocus();
  w_searchBox->selectAll();
  return;
}

// Insert word in the search box
void MainWindow::insertInSearchBox(const QString &word)
{
  w_searchBox->setText(word);
  return;
}

void MainWindow::clearBrowser(void)
{
  m_strBrowserHtml = "";
  w_browser->clear();
  return;
}

void MainWindow::setBrowserHtml(const QString &str, bool prs)
{
  w_browser->setHtml(m_strBrowserHtml + str);

  if (prs) { // persist.
    m_strBrowserHtml += str;
  }

  return;
}

// This slot is linked to the update button. Connect to the server and get updates.
void MainWindow::update(void)
{
  w_searchBox->clear();
  w_historyListBox->clearSelection();
  clearBrowser();

  if (m_strCurrentWord.length() > 0) {
    if (m_lstBackwardWords.count() >= MAX_HISTORY_SIZE) {
      m_lstBackwardWords.pop_front();
    }

    m_lstBackwardWords.push_back(m_strCurrentWord);
    m_lstForwardWords.clear();
    m_strCurrentWord = "";
  }

  genericUpdate();
  return;
}

// Slot linked to m_connectionTimer.
void MainWindow::abortUpdate(void)
{
  if (QHttp::Unconnected != m_http->state()  &&  QHttp::Closing != m_http->state()) {
    m_updateError = UPDATE_ERR_ABORTED;
    m_http->abort();
  }

  return;
}

// Slot linked to the m_connectionTimer.
void MainWindow::connectionTimeout(void)
{
  m_updateError = UPDATE_ERR_TIMEOUT;
  m_http->abort();
  return;
}

// Slot linked to QHttp::done.
void MainWindow::closeConnection(bool error)
{
  QString wc;
  m_connectionTimer.stop();

  // Close the database.
  dbClose();

  if (UPDATE_ERR_NO_ERROR == m_updateError) {
    error = false;
  }

  if (!error && -1 == m_newDefCountPending) {
    m_updateError = UPDATE_ERR_CONTENT;
    error = true;
  }

  if (error) {
    // Treat some special cases.
    if (UPDATE_ERR_TIMEOUT == m_updateError || UPDATE_ERR_ABORTED == m_updateError) {
      // "Minor" error. Recover gracefully below.
      error = false;

    } else if (!m_updateError) {
      // We never got a chance to set m_updateError, probably due to a network error.
      m_updateError = UPDATE_ERR_NETWORK;
    }
  }

  if (UPDATE_ERR_DB_READ != m_updateError  &&  UPDATE_ERR_DB_WRITE != m_updateError) {
    m_lastUpdateTime = m_lastUpdateTimePending;

    if (dbOpenWrite()) {
      QString luts = QString("%1").arg(m_lastUpdateTime);

      // Store the new update time in the database.
      if (!dbPut(LAST_UPDATE_TIME_KEY, luts.toAscii().data())) {
        logInfo("Error: new update time was not saved.");
        m_updateError = UPDATE_ERR_DB_WRITE;
        error = true;
      }

      dbClose();
    }
  }

  if (!error) {
    wc = QString::fromUtf8(" Terminat. %1 definiții noi sau actualizate.<br>").arg(m_newDefCount);

    if (0 != m_newDefCount) {
      wc += QString::fromUtf8(
              "Cele mai noi definiții pot fi găsite <u><a href=\"!nou\">aici</a></u> (maxim %1).<br>"
            )
            .arg(MAX_NEW_DEFINITIONS_DISPLAYED);
    }

    if (m_autoUpdate && 0 == w_historyListBox->count()) {
      wc += QString::fromUtf8(
              "<br>Baza de date se actualizează automat la pornirea programului. Pentru a schimba "
              "această preferință apăsați <u><a href=\"!lazimanual\">aici</a></u>. Actualizarea "
              "se poate face în orice moment folosind butonul de actualizare.<br>"
            );
    }

    setBrowserHtml(wc);

    if (UPDATE_ERR_TIMEOUT == m_updateError
        || UPDATE_ERR_NOT_DONE == m_updateError
        || UPDATE_ERR_ABORTED == m_updateError) {
      logInfo("Warning: Connection timeout or abort. Download cut short.");
      wc = QString::fromUtf8(
             "<font color=\"#993300\">Legătura s-a întrerupt prematur. "
             "Apăsați butonul de actualizare pentru a descărca restul de definiții.</font><br>"
           );
      logInfo(QString("Debug: Last Update Time: %1").arg(m_lastUpdateTime));
      setBrowserHtml(wc);

    } else {
      logInfo("Debug: Finished all data from server.");
    }

  } else { // if an error
    logInfo(QString::fromUtf8("Warning: Error during update: %1")
            .arg(UPDATE_ERRORS_STRING[m_updateError]));
    QString wc = QString::fromUtf8(
                   "<font color='#993300'>Eroare (%1).</font><br> Data ultimei "
                   "actualizări (sau a ultimului cuvânt actualizat): %2.<br>"
                 )
                 .arg(QString::fromUtf8(UPDATE_ERRORS_STRING[m_updateError]))
                 .arg(formatRoDate(m_lastUpdateTime));

    if (UPDATE_ERR_DB_READ == m_updateError || UPDATE_ERR_DB_WRITE == m_updateError) {
      wc += QString::fromUtf8("Eroare în baza de date.");
    }

    setBrowserHtml(wc);
  }

  // Build an in-memory word list.
  if (0 != m_newDefCount) {
    m_pThread->start();
  }

  w_browser->setEnabled(true);
  w_historyListBox->setEnabled(true);
  w_toolBar->setEnabled(true);
  setFocusToSearchBox();
  return;
}

// This slot is linked to the QHttp::responseHeaderReceived signal.
void MainWindow::checkHttpResponseHeader(const QHttpResponseHeader &rh)
{
  if (200 != rh.statusCode()) {
    logInfo(QString("Error: HTTP Server returned status code %1 %2")
            .arg(rh.statusCode())
            .arg(rh.reasonPhrase()));
    m_updateError = UPDATE_ERR_SERVER;
    m_http->abort();

  } else if (!(rh.contentType().contains("gzip")
               || rh.value(QString::fromAscii("content-encoding")).contains("gzip"))) {
    logInfo(QString("Error: Unknown content type received from server: %1")
            .arg(rh.contentType()));
    m_updateError = UPDATE_ERR_CONTENT;
    m_http->abort();
  }

  return;
}

// This slot is linked to the QHttp::readyRead signal.
void MainWindow::getDexonlineData(const QHttpResponseHeader &rh)
{
  m_connectionTimer.setSingleShot(true);
  m_connectionTimer.start(20 * 1000);

  static QByteArray xmlData;
  static bool gzipped = false;

  if (m_activeConnection) {
    m_newDefCount = 0;
    m_newDefCountPending = -1;

    if (200 != rh.statusCode()) {
      logInfo(QString("Error: HTTP Server returned status code %1 %2")
              .arg(rh.statusCode())
              .arg(rh.reasonPhrase()));
      m_updateError = UPDATE_ERR_SERVER;
      m_http->abort();
      return;
    }

    // Initialize the zlib inflate structures, if necessary.
    if (rh.contentType().contains("gzip")
        || rh.value(QString::fromAscii("content-encoding")).contains("gzip")) {
      gzipped = true;
      inflateData(xmlData, NULL, 0, true);

    } else {
      logInfo(QString("Error: Unknown content type received from server: %1")
              .arg(rh.contentType()));
      m_updateError = UPDATE_ERR_CONTENT;
      m_http->abort();
      return;
    }

    xmlData.truncate(0);
    m_activeConnection = false;
  }

  int datalen = xmlData.length();

  if (datalen < 1000000) {
    int basz = m_http->bytesAvailable();

    if (!basz) {
      logInfo("Error: getDexonlineData called, but no XML data is available.");
      return;
    }

    // Don't read more than this at a time.
    if (basz > 1000000) {
      basz = 1000000;
    }

    char *buf = new char[basz + 1];
    int buf_size = m_http->read(buf, basz);

    if (-1 == buf_size) {
      logInfo("Error occured during download (readBlock returned -1).");
      m_updateError = UPDATE_ERR_STREAM;
      m_http->abort();
      delete [] buf;
      return;
    }

    // Inflate, if needed.
    if (gzipped) {
      inflateData(xmlData, buf, buf_size, false);

    } else {
#if QT_VERSION >= 0x040500
      xmlData.append(buf, buf_size);
#else
      appendToByteArray(xmlData, buf, buf_size);
#endif
    }

    delete [] buf;
  }

  // Process XML data when 900k chars are available.
  datalen = xmlData.length();
  bool bytesAvailable = (bool) m_http->bytesAvailable();

  if (datalen >= 900000 || !(bytesAvailable)) {
    bool abortHttp = false;
    int lastidx = dbSaveXMLData(xmlData, &m_updateError, &abortHttp);

    if (abortHttp) {
      m_http->abort();
      return;
    }

    // If anything processed, discard it.
    if (0 != lastidx) {
      xmlData = xmlData.mid(lastidx, datalen);
    }

    QString wc = QString::fromUtf8(
                   " <b>%1/%2</b> definiții noi sau actualizate (apăsați <i>Esc</i> pentru a întrerupe)."
                 )
                 .arg(m_newDefCount)
                 .arg(m_newDefCountPending);
    setBrowserHtml(wc, false);
  }

  if (bytesAvailable) {
    emit moreData(rh);
  }

  return;
}

// Start a connection with the server.
void MainWindow::genericUpdate(void)
{
  if (m_pThread) {
    m_pThread->wait();
  }

  QString lbuf = QString::fromUtf8(
                   "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\">"
                   "<tr><td bgcolor=\"#eeeeee\"><b>Actualizare</b></td></tr><tr><td>"
                   "Programul încearcă să contacteze serverul %1... "
                 )
                 .arg(DEFAULT_SERVER_NAME);
  setBrowserHtml(lbuf);
  w_searchBox->clearFocus();
  w_toolBar->setEnabled(false);
  w_browser->setEnabled(false);
  w_historyListBox->setEnabled(false);

  m_updateError = UPDATE_ERR_NO_ERROR;
  m_lastUpdateTimePending = m_lastUpdateTime;

  QString tmpStr;
  tmpStr = HTTP_GET_REQUEST;
  QHttpRequestHeader header("GET", tmpStr.arg(m_lastUpdateTime));
  header.setValue("Host", DEFAULT_SERVER_NAME);

  tmpStr = USER_AGENT;
  header.setValue("User-Agent", tmpStr.arg(VERSION));
  header.setValue("Accept-Encoding", "gzip");
  header.setValue("Connection", "Keep-Alive");

  // Timeout after 20 seconds.
  m_connectionTimer.setSingleShot(true);
  m_connectionTimer.start(20 * 1000);
  m_activeConnection = true;
  m_newDefCount = 0;
  m_newDefCountPending = -1;

  m_http->setHost(DEFAULT_SERVER_NAME);
  m_http->request(header);
  return;
}

// Implements the Levenshtein distance algorithm found at http://www.merriampark.com/ldcpp.htm
// If 0 == maxdist, return the true Levenshtein (edit) distance. Otherwise, return the minimum
// between maxdist and the edit distance. If source or target are longer than 50, returns -1.
int MainWindow::LevenshteinDistance(const QString &source, const QString &target, int maxdist)
{
  char n, m;
  int l;

  l = source.length();

  if (l > 50) {
    return -1;
  }

  n = l;

  l = target.length();

  if (l > 50) {
    return -1;
  }

  m = l;

  if (0 == maxdist) {
    maxdist = qMax(m, n);
  }

  if (0 == n) {
    return qMin(int(m), maxdist);
  }

  if (0 == m) {
    return qMin(int(n), maxdist);
  }

  // Store the min. value on each column, so that,
  // if it reaches maxdist, we break early.
  char mincolval;

  char matrix[51][51];

  char j;
  char i;
  char cell;

  for (j = 0; j <= m; j++) {
    matrix[0][int(j)] = j;
  }

  for (i = 1; i <= n; i++) {
    mincolval = qMax(m, i);
    matrix[int(i)][0] = i;

    const QChar s_i = source[i-1];

    for (j = 1; j <= m; j++) {
      const QChar t_j = target[j-1];
      char  cost = (s_i == t_j ? 0 : 1);
      const char above = matrix[int(i)-1][int(j)];
      const char left = matrix[int(i)][int(j)-1];
      const char diag = matrix[int(i)-1][int(j)-1];
      cell = qMin(above + 1, qMin(left + 1, diag + cost));

      // Cover transposition, in addition to deletion, insertion and substitution.
      // This step is taken from Hal Berghel and David Roach:
      // "An Extension of Ukkonen's Enhanced Dynamic Programming ASM Algorithm"
      // http://www.berghel.net/publications/asm/asm.php

      if (i > 2 && j > 2) {
        char trans = matrix[int(i)-2][int(j)-2] + 1;

        if (source[i-2] != t_j) {
          trans++;
        }

        if (s_i != target[j-2]) {
          trans++;
        }

        if (cell > trans) {
          cell = trans;
        }
      }

      mincolval = qMin(mincolval, cell);
      matrix[int(i)][int(j)] = cell;
    }

    if (mincolval >= maxdist) {
      break;
    }
  }

  if (i == n + 1) {
    return int(matrix[int(n)][int(m)]);

  } else {
    return maxdist;
  }
}

void MainWindow::logInfo(const QString msg)
{
  if (m_logInfo) {
    if (m_pThread) {
      m_pThread->wait();
    }

    qWarning("%s", msg.toAscii().data());
  }

  return;
}

QString MainWindow::formatRoDate(int seconds, bool hour)
{
  if (0 == seconds) {
    return QString::fromUtf8("Niciodată");
  }

  const char *MONTH_NAMES[] = {
    "ianuarie", "februarie", "martie", "aprilie", "mai", "iunie",
    "iulie", "august", "septembrie", "octombrie", "noiembrie", "decembrie"
  };

  int month;

  QDateTime qtTime = QDateTime::fromTime_t(seconds);

  month = qtTime.toString("M").toInt() - 1;

  QString result;

  if (hour) {
    result = QString("%1%2%3")
             .arg(qtTime.toString("d "))
             .arg(MONTH_NAMES[month])
             .arg(qtTime.toString(" yyyy, h:mm"));

  } else {
    result = QString("%1%2%3")
             .arg(qtTime.toString("d "))
             .arg(MONTH_NAMES[month])
             .arg(qtTime.toString(" yyyy"));
  }

  return result;
}

QString MainWindow::processCommands(const QString &cmd)
{
  QString result;

  if (cmd.startsWith("!surs") || cmd.startsWith("!dex") || "!dar" == cmd
      || "!der" == cmd  || "!dge" == cmd || "!dlra" == cmd || "!dlrlv" == cmd
      || "!dlrlc" == cmd || "!dor" == cmd || "!dlrc" == cmd || "!dlrm" == cmd
      || "!dmlr" == cmd || "!doom" == cmd || "!mda" == cmd || "!mdn" == cmd
      || "!neoficial" == cmd || cmd.startsWith("!anton")) {
    result = loadFile(":/html/sources.html");

  } else if ("!info" == cmd || "!despre" == cmd) {
    result = QString::fromUtf8(
               "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
               "<b>Dixit</b></td></tr><tr><td><p align=\"justify\">Dixit permite căutări de cuvinte "
               "<i>offline</i> folosind definițiile adunate de proiectul <i>DEX online</i>. Baza de date "
               "locală conține %1 definiții culese din diverse <u><a href=\"!surse\">surse tipărite "
               "sau orale</a></u>. Data ultimei actualizări: %2.<br></p></td></tr>"
             )
             .arg(m_totalDefinitions)
             .arg(formatRoDate(m_lastUpdateTime, true))

#if defined(Q_CC_MSVC)
             + loadFile(":/html/lastinfow.html");
#else
             + loadFile(":/html/lastinfo.html");
#endif

  } else if ("!instructiuni" == cmd || "!ajutor" == cmd || "!help" == cmd) {
    result = loadFile(":/html/help.html")
             + QString::fromUtf8(
               "<p align=\"justify\">Baza de date conținând definițiile este păstrată local și %1 se "
               "actualizează la pornirea programului%2. Pentru a schimba această preferință apăsați "
               "<u><a href=\"!lazi%3\">aici</a></u>. Actualizarea se poate face în orice moment folosind "
               "butonul de actualizare.</p>"
             )
             .arg(m_autoUpdate ? "" : "nu")
             .arg(m_autoUpdate ? QString::fromUtf8(" dacă există o legătură la Internet ") : "")
             .arg(m_autoUpdate ? "manual" : "auto")

             + loadFile(":/html/cmd.html")

             + QString::fromUtf8(
               "<p align=\"justify\">Programul %1 afișează informațiile interne referitoare la execuția "
               "programului. Pentru a schimba această preferință apăsați <u><a href=\"!loginfo\">aici"
               "</a></u>.</p></td></tr></table>"
             )
             .arg(m_logInfo ? "" : "nu");

  } else if (cmd.startsWith("!lazi")) {
    if (cmd.startsWith("!laziauto")) {
      m_autoUpdate = true;

    } else if (cmd.startsWith("!lazimanu")) {
      m_autoUpdate = false;
    }

    result = QString::fromUtf8(
               "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
               "<b>Modul de actualizare a fost schimbat</b></td></tr><tr><td>Lista de definiții <b>%1 se "
               "actualizează automat</b> la pornirea programului. Pentru a schimba această preferință "
               "apăsați <u><a href=\"!lazi%2\">aici</a></u>.</td></tr></table>"
             )
             .arg(m_autoUpdate ? "" : "nu")
             .arg(m_autoUpdate ? "manual" : "auto");

  } else if (cmd.startsWith("!abr") || cmd.startsWith("!presc")) {
    result = loadFile(":/html/abrv.html");

  } else if ("!nou" == cmd || "!recent" == cmd) {
    result = QString::fromUtf8(
               "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
               "<b>Cele mai noi definiții (maxim %1)</b></td></tr>"
             )
             .arg(MAX_NEW_DEFINITIONS_DISPLAYED);

    if (m_lstNewDefinitions.isEmpty()) {
      result += QString::fromUtf8(
                  "<tr><td><p>Nimic nou. Data ultimei actualizări: %1.</p><br></td></tr></table>"
                )
                .arg(formatRoDate(m_lastUpdateTime));

    } else {
      result += QString::fromUtf8("<tr><td align=\"right\"><font size=-1 color='#686868'>"
                                  "<a href=\"!abrevieri\">Abrevieri</a></font></td></tr></table>");
      int m = 0;

      for (QStringList::iterator it = m_lstNewDefinitions.begin(); it != m_lstNewDefinitions.end(); it++) {
        if (m > MAX_NEW_DEFINITIONS_DISPLAYED) {
          break;
        }

        result += htmlizeBrowserEntry(*it);
        m++;
      }
    }

  } else if ("!gplro" == cmd) {
    result = loadFile(":/html/gplro.html");

  } else if ("!gpl" == cmd) {
    result = loadFile(":/html/gpl.html");

  } else if ("!dixit" == cmd) {
    result = loadFile(":/html/dixit.html");

  } else if ("!pp" == cmd) {
    result = loadFile(":/html/fp.html");

  } else if ("!loginfo" == cmd) {
    m_logInfo = !m_logInfo;
    result = QString::fromUtf8(
               "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
               "<b>Informații specifice execuției programului</b></td></tr><tr><td> Informațiile "
               "specifice execuției programului <b>%1 sunt afișate</b>. Pentru a schimba această "
               "preferință apăsați <u><a href=\"!loginfo\">aici</a></u>.</td></tr></table>"
             )
             .arg(m_logInfo ? "" : "nu");

  } else {
    // unknown command
    result = QString::fromUtf8(
               "<table cellpadding=\"0\" cellspacing=\"2\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
               "<b>Programul nu poate procesa comanda </b></td></tr><tr><td>Apăsați <u><a href="
               "\"!instructiuni\">aici</a></u> pentru a consulta pagina de instrucțiuni.</td></tr></table>"
             );
  }

  return result;
}

#if QT_VERSION < 0x040500
void MainWindow::appendToByteArray(QByteArray &ba, const char  *buf, uint bufsz)
{
  uint oldsz = ba.size();
  ba.resize(oldsz + bufsz);
  memcpy(ba.begin() + oldsz, buf, bufsz);
  return;
}
#endif

// Process zdata and append inflated result to idata.
void MainWindow::inflateData(QByteArray &idata, char *zdata, int zdatasz, bool init_only)
{
  int error, len;
  bool done = false;
  static z_stream gz_stream;
  char gzbuffer[65536];

  if (init_only) {
    // Initialize zlib inflate.
    gz_stream.zalloc = NULL;
    gz_stream.zfree = NULL;
    gz_stream.opaque = NULL;
    gz_stream.next_in = gz_stream.next_out = 0;
    gz_stream.avail_in = gz_stream.avail_out = 0;

    if (Z_OK != inflateInit2(&gz_stream, 16 + 15)) {
      logInfo("Error initializing gzip inflate. (zlib 1.2 needed)");
    }

    return;
  }

  gz_stream.next_out = (Bytef *) gzbuffer;
  gz_stream.avail_out = sizeof(gzbuffer);
  gz_stream.next_in = (Bytef *) zdata;
  gz_stream.avail_in = zdatasz;

  while (0 != gz_stream.avail_in) {
    if (0 == gz_stream.avail_out) {
#if QT_VERSION >= 0x040500
      idata.append(gzbuffer, sizeof(gzbuffer));
#else
      appendToByteArray(idata, gzbuffer, sizeof(gzbuffer));
#endif
      gz_stream.next_out = (Bytef *) gzbuffer;
      gz_stream.avail_out = sizeof(gzbuffer);
    }

    error = inflate(&gz_stream, Z_NO_FLUSH);

    if (Z_STREAM_END == error) {
      break;

    } else if (Z_OK != error) {
      logInfo("Error during inflate. Stream is not gzipped or zlib version is too old (zlib 1.2 needed).");
      exit(1);
    }
  }

  while (true) {
    len = sizeof(gzbuffer) - gz_stream.avail_out;

    if (len > 0) {
#if QT_VERSION >= 0x040500
      idata.append(gzbuffer, len);
#else
      appendToByteArray(idata, gzbuffer, len);
#endif
      gz_stream.next_out = (Bytef *) gzbuffer;
      gz_stream.avail_out = sizeof(gzbuffer);
    }

    if (done) {
      break;
    }

    error = inflate(&gz_stream, Z_SYNC_FLUSH);

    if (0 == len  &&  Z_BUF_ERROR == error) {
      error = Z_OK;
    }

    done = (0 != gz_stream.avail_out  ||  Z_STREAM_END == error);

    if (Z_OK != error  &&  Z_STREAM_END != error) {
      break;
    }
  }

  return;
}

// Replace #, @, $, %, |, ^, _ with HTML tags, and some
// dexonline scripting (" * ", " ** ", " - ", "") with unicode
void MainWindow::htmlizeDefinition(QString &html, const char *in)
{
  if (0 == qstrlen(in)) {
    return;
  }

  bool in_abr = false, in_bold = false, in_italic = false, in_quote = false;
  QChar h_i;
  int i = 0, m, n;
  QString text, link, html_expr;
  html = QString::fromUtf8(in);

  do {
    h_i = html[i];

    if ('#' == h_i) { // #abr.#
      if ('&' != html[i-1]) { // don't replace unicode chars (e.g. &#x2013;)
        html.replace(i, 1, (in_abr ? "</font>" : "<font color=DarkViolet>"));
        in_abr = !in_abr;
      }

    } else if ('@' == h_i) { // @bold@
      if ('\\' != html[i-1]) { // don't replace escaped @ (\@)
        html.replace(i, 1, (in_bold ? "</b>" : "<b>"));
        in_bold = !in_bold;
      }

    } else if ('$' == h_i) { // $italic$
      if ('\\' != html[i-1]) { // don't replace escaped $ (\$)
        html.replace(i, 1, (in_italic ? "</i>" : "<i>"));
        in_italic = !in_italic;
      }

    } else if ('"' == h_i) { // "quote"
      if (!('=' == html[i-1] || '>' == html[i+1])) { // don't replace html links
        html.replace(i, 1, (in_quote ? "&#8220;" : "&#8222;"));
        in_quote = !in_quote;
      }

    } else if ('%' == h_i) { // %aerisit%
      if ('\\' != html[i-1]) { // don't replace escaped % (\%)
        m = html.indexOf("%", i + 1); // position of the 2nd %

        if (-1 != m) {
          text = html.mid(i + 1, m - i - 1);
          n = text.length();

          if (n < 20) { // In DE there are no escaped % (\%)
            html_expr = "";

            for (m = 0; m < n; m++) {
              if (m != n - 1) {
                html_expr += text[m] + " ";

              } else {
                html_expr += text[m];
              }
            }

            html.replace(i, n + 2, html_expr);
          }
        }
      }

    } else if ('|' == h_i) { // |text|link|
      m = html.indexOf("|", i + 1); // position of the 2nd bar
      n = html.indexOf("|", m + 1); // position of the 3rd bar

      if (-1 != m  &&  -1 != n) {
        text = html.mid(i + 1, m - i - 1);
        link = html.mid(m + 1, n - m - 1);
        html_expr = "<a href=\"" + link + "\">" + text + "</a>";
        html.replace(i, text.length() + link.length() + 3, html_expr);
      }

    } else if ('^' == h_i) { // <sup></sup>
      if ('{' != html[i+1]) {
        text = html.mid(i + 1, 1);
        html_expr = "<sup>" + text + "</sup>";
        html.replace(i, 2, html_expr);

      } else {
        m = html.indexOf("}", i + 2);

        if (-1 != m) {
          text = html.mid(i + 2, m - i - 2);
          html_expr = "<sup>" + text + "</sup>";
          html.replace(i, text.length() + 3, html_expr);
        }
      }

    } else if ('_' == h_i) { // <sub></sub>
      if ('{' != html[i+1]) {
        text = html.mid(i + 1, 1);
        html_expr = "<sub>" + text + "</sub>";
        html.replace(i, 2, html_expr);

      } else {
        m = html.indexOf("}", i + 2);

        if (-1 != m) {
          text = html.mid(i + 2, m - i - 2);
          html_expr = "<sub>" + text + "</sub>";
          html.replace(i, text.length() + 3, html_expr);
        }
      }

    } else if ('*' == h_i) {
      if ('*' == html[i+1]) {
        html.replace(i, 2, "&diams;"); // **

      } else {
        html.replace(i, 1, "&loz;"); // *
      }

    } else if ('-' == h_i) { // put a longer -
      if (' ' == html[i-1] && ' ' == html[i+1]) {
        html.replace(i - 1, 3, " &#x2013; "); // put a longer -
      }
    }

    i++;
  } while (i < html.length());

  return;
}

QString MainWindow::getDictionaryName(const QString &shortName)
{
  QString result;
  int row;

  for (row = 1; row < DICTIONARIES_COUNT; row++) {
    if (DICTIONARIES[row][0] == shortName) {
      result = QString("<a href=\"!surse\">%1</a>").arg(QString::fromUtf8(DICTIONARIES[row][1]));
      break;
    }
  }

  // If dictionary name is unknown, take its name from XML, and don't add a link for it
  if (DICTIONARIES_COUNT == row) {
    result = shortName;
  }

  return result;
}

// Htmlize an entry for the w_browser.
QString MainWindow::htmlizeBrowserEntry(const QString &dbRec)
{
  int idx0 = 0, idx1;
  int datalen = dbRec.length();

  for (idx1 = 0; idx1 < datalen; idx1++) {
    if ('<' == dbRec[idx1]) {
      idx0 = idx1;
      continue;
    }

    if ('>' == dbRec[idx1]) {
      break;
    }
  }

  QString sbuf;
  htmlizeDefinition(sbuf,  dbRec.mid(idx1 + 1, datalen - idx1 - 1).toAscii().data());
  QString result = "<table cellpadding=0 cellspacing=3 width=\"100%\"><tr><td>";
  result += sbuf;
  result += "<br><font size=-1 color='#708090'>Sursa: ";
  result += getDictionaryName(dbRec.mid(idx0 + 1, idx1 - idx0 - 1));
  result += "</font></td></tr></table>";
  return result;
}

// This slot is linked to bigger Font action (Alt + +)
void MainWindow::makeFontBigger(void)
{
  QFont newFont = font();
  int newsz = newFont.pointSize() + 1;

  if (newsz > 16) {
    newsz = 16;
  }

  newFont.setPointSize(newsz);
  setFont(newFont);
  return;
}

// This slot is linked to smaller Font action (Alt + -)
void MainWindow::makeFontSmaller(void)
{
  QFont newFont = font();
  int newsz = newFont.pointSize() - 1;

  if (newsz < 8) {
    newsz = 8;
  }

  newFont.setPointSize(newsz);
  setFont(newFont);
  return;
}

// This slot is linked to change Font action (Alt + F)
void MainWindow::changeFont(void)
{
  bool ok;
  QFont newFont = QFontDialog::getFont(&ok, font(), this);

  if (ok) {
    if (newFont.pointSize() > 16) {
      newFont.setPointSize(16);

    } else if (newFont.pointSize() < 8) {
      newFont.setPointSize(8);
    }

    setFont(newFont);
  }

  return;
}

// Called from the constructor.
void MainWindow::dbStartIndexing(void)
{
  clearBrowser();
  w_historyListBox->clear();

  // dbSetup returns true if an update is required
  if (dbSetup()) {
    genericUpdate();

  } else {
    // Build an in-memory word index, asynchronously.
    m_pThread->start();
    setBrowserHtml(processCommands("!pp"));
  }

  setFocusToSearchBox();
  return;
}

// Return true if a database update is required, otherwise return false
bool MainWindow::dbSetup(void)
{
  int iarg = 1;

  // Jump over potential command line options.
  while (qApp->argc() > iarg && qApp->argv()[iarg][0] == '-') {
    iarg++;
  }

  QDir dbd; // Current directory.
  dbd.makeAbsolute();

  // Get full path for db file.
  if (qApp->argc() > iarg) {
    qstrncpy(m_dbFileName, QDir::cleanPath(dbd.filePath(qApp->argv()[iarg])).toAscii(), 128);

  } else {
    qstrncpy(m_dbFileName, dbd.filePath(DB_FILE_NAME).toAscii(), 128);
  }

  m_pThread = new dbIndexThread(this);
  logInfo(QString("Debug: Database file: %1").arg(m_dbFileName));

  // Find last update time under <> key in the database.
  bool no_database = false;

  if (!dbOpenRead()) {
    // Probably a new installation.
    no_database = true;
    m_lastUpdateTime = LAST_UPDATE_TIME_BASE;
    logInfo("Info: The program will create a new database.");

  } else {
    bool ok;
    char *dbLastUpdateTime;
    dbLastUpdateTime = dbGet(LAST_UPDATE_TIME_KEY, &ok);

    if (ok) {
      m_lastUpdateTime = QByteArray(dbLastUpdateTime).toInt();
      free(dbLastUpdateTime);

    } else {
      logInfo("Error: No timestamp was found in the database.");
      m_lastUpdateTime = LAST_UPDATE_TIME_BASE;
    }
  }

  dbClose();
  return(m_autoUpdate || no_database);
}

// Save XML data into qdbm. Return the number of chars processed.
int MainWindow::dbSaveXMLData(const QString &xmlData, UpdateErrors *updateError, bool *abortHttp)
{
  // This is the last update time advertised by the server at the beginning
  // of the transmition. It is used only when all data was received okay.
  // Otherwise, the individual time in each entry is taken.
  static time_t lut_advertised = 0;

  QChar nextch1, nextch2;
  int   lastIdx = 0, idx = 0;
  int   oIdx[6] = {0, 0, 0, 0, 0, 0};
  int   cIdx[6] = {0, 0, 0, 0, 0, 0};
  QString dbNewWord, dbNewRec, entryId, entryWord, entryTimeStamp;

  *updateError = UPDATE_ERR_NOT_DONE;
  int datalen  = xmlData.length();

  while (idx < datalen) {

    // look only for tags
    if ('<' != xmlData[idx]) {
      idx++;
      continue;
    }

    nextch1 = xmlData[idx+1];
    nextch2 = xmlData[idx+2];

    // Check if we haven't seen an entry yet.
    if (0 == oIdx[0]) {
      // Test for the update timestamp.
      if ('T' == nextch1) {
        idx += 11;
        oIdx[2] = idx;
        continue;

      } else if ('/' == nextch1 && 'T' == nextch2) {
        cIdx[2] = idx;
        idx += 12;
        lut_advertised = xmlData.mid(oIdx[2], cIdx[2] - oIdx[2]).toInt();
        continue;

      } else if ('N' == nextch1 && 'u' == nextch2) {
        idx += 12;
        oIdx[5] = idx;
        continue;

      } else if ('/' == nextch1 && 'N' == nextch2 && 'u' == xmlData[idx+3]) {
        cIdx[5] = idx;
        m_newDefCountPending = xmlData.mid(oIdx[5], cIdx[5] - oIdx[5]).toInt();

        if (m_newDefCountPending) {
          // Open the database.
          if (!dbOpenWrite()) {
            *updateError = UPDATE_ERR_DB_WRITE;
            *abortHttp = true;
            return 0;
          }
        }

        idx += 13;
        continue;
      }
    }

    if ('/' == nextch1  &&  'D' == nextch2  &&  'i' == xmlData[idx+3]) {
      if (lut_advertised > 0) {
        m_lastUpdateTimePending = lut_advertised;
      }

      dbClose();
      *updateError = UPDATE_ERR_NO_ERROR;
      *abortHttp = true;
      idx += 13;

    } else if ('T' == nextch1) {
      idx += 11;
      oIdx[2] = idx;

    } else if ('/' == nextch1 && 'T' == nextch2) {
      cIdx[2] = idx;
      idx += 12;

    } else if ('S' == nextch1) {
      idx += 8;
      oIdx[3] = idx;

    } else if ('/' == nextch1 && 'S' == nextch2) {
      cIdx[3] = idx;
      idx += 9;

    } else if ('D' == nextch1 && 'e' == nextch2) {
      idx += 12;
      oIdx[4] = idx;

    } else if ('/' == nextch1 && 'D' == nextch2 && 'e' == xmlData[idx+3]) {
      cIdx[4] = idx;
      idx += 13;

    } else if ('N' == nextch1 && 'a' == nextch2) {
      idx += 6;
      oIdx[1] = idx;

    } else if ('/' == nextch1 && 'N' == nextch2 && 'a' == xmlData[idx+3]) {
      cIdx[1] = idx;
      idx += 7;

    } else if ('I' == nextch1 && 'd' == nextch2) {
      idx += 4;
      oIdx[0] = idx;

    } else if ('/' == nextch1 && 'I' == nextch2 && 'd' == xmlData[idx+3]) {
      cIdx[0] = idx;
      idx += 5;

      // have a record, update the db

    } else if ('/' == nextch1   &&  'E' == nextch2  &&  'n' == xmlData[idx+3]) {

      if (-1 == m_newDefCountPending) {
        *updateError = UPDATE_ERR_CONTENT;
        *abortHttp = true;
        return 0;
      }

      if (!m_dbHandle) {
        logInfo("Internal Error: Database is not open during update.");
        *updateError = UPDATE_ERR_DB_WRITE;
        *abortHttp = true;
        return 0;
      }

      entryId        = xmlData.mid(oIdx[0], cIdx[0] - oIdx[0]);
      entryWord      = xmlData.mid(oIdx[1], cIdx[1] - oIdx[1]);
      entryTimeStamp = xmlData.mid(oIdx[2], cIdx[2] - oIdx[2]);

      dbNewRec = QString("%1<%2>%3")
                 .arg(entryId)
                 .arg(xmlData.mid(oIdx[3], cIdx[3] - oIdx[3]))
                 .arg(xmlData.mid(oIdx[4], cIdx[4] - oIdx[4]));

      bool  ok = true, db_error = false, sameKey;
      char *dupe;
      int   ij = 0, keysz;
      dbNewWord = entryWord;

      while (dbKeyExists(dbNewWord.toAscii().data())) {
        // Check whether this entry word exists with the same Id.
        dupe = dbGet(dbNewWord.toAscii().data(), &ok);

        if (ok) {
          sameKey = true;
          keysz = entryId.length();

          if ('<' != dupe[keysz]) {
            sameKey = false;

          } else {
            for (int k = 0; k < keysz; k++) {
              if (entryId[k] != dupe[k]) {
                sameKey = false;
                break;
              }
            }
          }

          free(dupe);

          if (sameKey) {
            break;
          }

        } else {
          logInfo(QString(
                    "Internal error: updated entry word \"%1\" exists, but cannot be retrieved"
                  )
                  .arg(entryWord));
          db_error = true;
        }

        dbNewWord = QString("%1%2").arg(entryWord).arg(ij);
        ij++;
      }

      if (db_error) {
        *updateError = UPDATE_ERR_DB_READ;
        *abortHttp = true;
        return 0;
      }

      if (!dbPut(dbNewWord.toAscii().data(), dbNewRec.toAscii().data())) {
        *updateError = UPDATE_ERR_DB_WRITE;
        *abortHttp = true;
        return 0;
      }

      m_lastUpdateTimePending = entryTimeStamp.toInt() - 1;
      m_newDefCount++;

      if (m_newDefCount + MAX_NEW_DEFINITIONS_DISPLAYED > m_newDefCountPending) {
        m_lstNewDefinitions.push_front(dbNewRec);

        if (m_lstNewDefinitions.size() > MAX_NEW_DEFINITIONS_DISPLAYED) {
          m_lstNewDefinitions.pop_back();
        }
      }

      idx += 8;
      lastIdx = idx;

    } else {
      idx++;
    }
  }

  return lastIdx;
}

QString MainWindow::dbResults(const QString &word)
{
  bool duplicates;
  int ij, cnt_dict_entries = 0, cnt_distinct_words = 0, cnt_exact_matches = 0;
  QString awij, buf, info, lword(word), results, sbuf;
  QStringList deflist, dictwordlist, wordlist;
  QStringList::iterator it;

  buf = QString::fromAscii("^(") + lword;
  buf.replace('*', "[ \\-a-z]*");
  buf.replace('?', "[ \\-a-z]");
  buf += QString::fromAscii(")$");

  // Case insensitive, need to get DE words (e.g. Adam)
  QRegExp rx = QRegExp(buf, Qt::CaseInsensitive, QRegExp::RegExp);
  rx.setMinimal(true);

  if (!dbOpenRead()) {
    results = QString::fromUtf8(
                "<font color='#993300'>Eroare: Programul nu poate citi din baza de date.</font><br> "
                "Verificați existența bazei de date (%1). Dacă există, s-ar putea să fie în curs de "
                "actualizare de către altă instanță a programului, sau s-ar putea să fie coruptă.<br>"
              )
              .arg(m_dbFileName);
    dbClose();
    return results;
  }

  // Find all matching words by searching in m_lstAllWords.
  for (it = m_lstAllWords.begin(); it != m_lstAllWords.end(); it++) {
    sbuf = *it;

    if (-1 != sbuf.indexOf(rx)) {
      wordlist.append(sbuf);
      // Check for other definitions for this word.
      // They are stored in the database with keys word0, word1, etc
      ij = 0;

      do {
        awij = QString("%1%2").arg(sbuf).arg(ij);
        duplicates = dbKeyExists(awij.toAscii().data());

        if (duplicates) {
          wordlist.append(awij);
        }

        ij++;
      } while (duplicates);
    }
  }

  cnt_exact_matches = wordlist.size();

  // If no exact matches, try to find close matches.
  if (!cnt_exact_matches) {
    int  edist;

    for (it = m_lstAllWords.begin(); it != m_lstAllWords.end(); it++) {
      sbuf = *it;

      if (lword.length() + 2 <= sbuf.length() || lword.length() >= sbuf.length() + 2) {
        continue;
      }

      edist = LevenshteinDistance(lword, sbuf, 2);

      if (edist < 2) {
        wordlist.append(sbuf);

        if (wordlist.size() > MAX_DEFINITIONS_DISPLAYED) {
          break;
        }
      }
    }
  }

  // Count distinct words in wordlist
  QSet<QString> wordset;

  for (it = wordlist.begin(); it != wordlist.end(); it++) {
    sbuf = *it;
    rx = QRegExp("[0-9]");

    if (-1 == sbuf.indexOf(rx)) {
      wordset.insert(sbuf.toLower());
    }
  }

  cnt_distinct_words = wordset.size();

  // Count dictionary entries and fetch definitions only for cnt_distinct_words = 1
  if (wordlist.size()) {
    bool    ok;
    char   *dbRec;
    QString dbRawRec, dictionary, htmlEntry;

    if (0 == m_defaultDictionary) {
      dictionary = QString("<");

    } else {
      dictionary = QString("<%1>").arg(DICTIONARIES[m_defaultDictionary][0]);
    }

    for (it = wordlist.begin(); it != wordlist.end() && deflist.size() <= MAX_DEFINITIONS_DISPLAYED; it++) {
      dbRec = dbGet((*it).toAscii().data(), &ok);

      if (ok) {
        dbRawRec = QString(dbRec);
        free(dbRec);

        if (-1 == dbRawRec.indexOf(dictionary)) {
          continue;

        } else {
          cnt_dict_entries++;
          dictwordlist.append(*it);
        }

        if (1 == cnt_distinct_words) {
          htmlEntry = htmlizeBrowserEntry(dbRawRec);
        }

      } else {
        htmlEntry = QString::fromUtf8(
                      "<p><font color='#993300'>Eroare internă: Programul "
                      "nu a găsit cuvântul %1 în baza de date.</font></p>"
                    )
                    .arg(*it);
      }

      if (1 == cnt_distinct_words) {
        deflist.append(htmlEntry);
      }
    }
  }

  dbClose();

  results = QString::fromUtf8(
              "<table cellpadding=\"0\" cellspacing=\"1\" width=\"100%\"><tr><td bgcolor=\"#eeeeee\">"
            );

  if (0 == cnt_distinct_words) {
    info = QString::fromUtf8(
             "Nici o definiție nu a fost găsită pentru «%1».</td></tr></table>"
           )
           .arg(word);
    results += info;

  } else if (1 == cnt_distinct_words && cnt_exact_matches) {
    QString lbuf;

    if (0 != m_defaultDictionary && deflist.size() < cnt_exact_matches) {
      if (deflist.size() <= MAX_DEFINITIONS_DISPLAYED) {
        lbuf = QString::fromUtf8(" din toate sursele (%1 din %2)")
               .arg(deflist.size())
               .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));

      } else {
        lbuf = QString::fromUtf8(" din toate sursele (mai mult de %1 din %2)")
               .arg(MAX_DEFINITIONS_DISPLAYED)
               .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
      }
    }

    results += QString::fromUtf8(
                 "%1 definiți%2%3 pentru «%4»:</td></tr><tr><td align=\"right\"><font size=-1 "
                 "color='#686868'><a href=\"!abr\">Abrevieri</a></font></td></tr></table>"
               )
               .arg(cnt_exact_matches)
               .arg(1 == cnt_exact_matches ? "e" : "i")
               .arg(lbuf).arg(word);

    if (deflist.size() <= MAX_DEFINITIONS_DISPLAYED) {
      for (it = deflist.begin(); it != deflist.end(); it++) {
        results += *it;
      }
    }

  } else if (cnt_dict_entries <= MAX_DEFINITIONS_DISPLAYED) {
    int cntDistinctDictWords = 0;
    buf = "<br><br>";
    rx = QRegExp("[0-9]");

    if (0 == m_defaultDictionary) {
      wordlist.sort();

      for (it = wordlist.begin(); it != wordlist.end(); it++) {
        sbuf = *it;

        if (-1 == sbuf.indexOf(rx)) {
          buf += "| <a href=\"" + sbuf + "\">" + sbuf + "</a> ";
        }
      }

    } else {
      dictwordlist.sort();
      QString prev_sbuf("");

      for (it = dictwordlist.begin(); it != dictwordlist.end(); it++) {
        sbuf = *it;
        sbuf.replace(rx, "");

        // suffixless word is in list OR suffixless word is not in list, but word_ij is
        if (sbuf == (*it) || (-1 == dictwordlist.indexOf(sbuf) && sbuf != prev_sbuf)) {
          cntDistinctDictWords++;
          buf += "| <a href=\"" + sbuf + "\">" + sbuf + "</a> ";
        }

        prev_sbuf = sbuf;
      }
    }

    sbuf = QString::fromUtf8(", %1 definiți%2 pentru %3 %4 fiind din %5")
           .arg(cnt_dict_entries)
           .arg(1 == cnt_dict_entries ? "e" : "i")
           .arg(cntDistinctDictWords)
           .arg(1 == cntDistinctDictWords ? QString::fromUtf8("cuvânt") : "cuvinte")
           .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
    results += QString::fromUtf8(
                 "Căutarea «%1» a găsit %2 definiți%3 pentru %4 %5 din toate sursele%6.</td></tr></table>"
               )
               .arg(word)
               .arg(wordlist.size())
               .arg(wordlist.size() == 1 ? "e" : "i")
               .arg(cnt_distinct_words)
               .arg(1 == cnt_distinct_words ? QString::fromUtf8("cuvânt") : "cuvinte")
               .arg(0 == m_defaultDictionary ? QString::fromUtf8("") : sbuf);
    results += buf + "|";

  } else if (cnt_dict_entries > MAX_DEFINITIONS_DISPLAYED) {
    sbuf = QString::fromUtf8(" (%1 definiți%2 din %3)")
           .arg(cnt_dict_entries)
           .arg(1 == cnt_dict_entries ? "e" : "i")
           .arg(QString::fromUtf8(DICTIONARIES[m_defaultDictionary][1]));
    info = QString::fromUtf8(
             "%1 rezultate pentru «%2»%3.</td></tr></table><p><font color=\"#993300\">"
             "Numărul maxim de definiții care pot fi afișate este %4.<br>Folosiți un "
             "criteriu de căutare mai selectiv.</font></p>"
           )
           .arg(wordlist.size())
           .arg(word)
           .arg(0 == m_defaultDictionary ? QString::fromUtf8("") : sbuf)
           .arg(MAX_DEFINITIONS_DISPLAYED);
    results += info;
  }

  return results;
}

// This function runs in its own thread
int MainWindow::dbCreateIndex(void)
{
  if (!dbOpenRead()) {
    m_updateError = UPDATE_ERR_DB_READ;
    return 0;
  }

  char *key;
  int keysz, count = 0;

  m_lstAllWords.clear();

  // initialize the iterator
  if (!dpiterinit(m_dbHandle)) {
    logInfo(QString("Error: cannot initialize the iterator: %1").arg(dperrmsg(dpecode)));
    m_updateError = UPDATE_ERR_DB_READ;
    return 0;
  }

  // scan with the iterator to get all keys.
  while (NULL != (key = dpiternext(m_dbHandle, NULL))) {
    count++;
    keysz = qstrlen(key);

    // Store each word only once in the m_lstAllWords list, ie skip duplicates.
    // (duplicate words have a number at the end in the database)
    if (!(key[keysz-1] >= '0' && key[keysz-1] <= '9')) {
      m_lstAllWords.append(key);
    }

    free(key);
  }

  keysz = 1;

  if (dbKeyExists(CORRELATION_KEY)) {
    bool ok = true;
    key = dbGet(CORRELATION_KEY, &ok);

    if (ok) {
      keysz = QByteArray(key).toInt();
      free(key);
    }
  }

  dbClose();
  count -= keysz;
  m_totalDefinitions = count;
  return count;
}

bool MainWindow::dbOpenRead(void)
{
  if (!(m_dbHandle = dpopen(m_dbFileName, DP_OREADER, -1))) {
    logInfo(QString("Error dpopen: cannot open database for reading: %1").arg(dperrmsg(dpecode)));
    return false;
  }

  return true;
}

bool MainWindow::dbOpenWrite(void)
{
  if (!(m_dbHandle = dpopen(m_dbFileName, DP_OWRITER | DP_OCREAT, -1))) {
    logInfo(QString("Error dpopen: cannot open database for writing: %1").arg(dperrmsg(dpecode)));
    return false;
  }

  return true;
}

bool MainWindow::dbClose(void)
{
  if (m_dbHandle) {
    if (!dpclose(m_dbHandle)) {
      logInfo(QString("Error dpclose: cannot close database: %1").arg(dperrmsg(dpecode)));
      m_dbHandle = NULL;
      return false;
    }
  }

  m_dbHandle = NULL;
  return true;
}

bool MainWindow::dbKeyExists(const char *key)
{
  if (-1 == dpvsiz(m_dbHandle, key, -1)) {
    return false;
  }

  return true;
}

bool MainWindow::dbPut(const char *key, const char *data)
{
  if (!dpput(m_dbHandle, key, -1, data, -1, DP_DOVER)) {
    logInfo(QString("Error dpput: cannot save database data: %1").arg(dperrmsg(dpecode)));
    return false;
  }

  return true;
}

char *MainWindow::dbGet(const char *key, bool *ok)
{
  char *data;

  if (NULL != (data = dpget(m_dbHandle, key, -1, 0, -1, NULL))) {
    *ok = true;

  } else {
    logInfo(QString("Error dpget: cannot get database data: %1").arg(dperrmsg(dpecode)));
    *ok = false;
  }

  return data;
}

// END OF FILE

